//
//  HTTPThread.cpp
//
//  An thread used to retrieve files from the Internet.
//

#define	STRICT

#include "stdafx.h"
#include "HThrd.h"

CHTTPThread::CHTTPThread()
{
    m_hSession = NULL ;
    m_pThread = NULL ;
}

BOOL CHTTPThread::IsThreadAvailable()
{
    return (m_pThread == NULL) ;
}

void CHTTPThread::ResetSession()
{
    if (m_hSession != NULL)
    {
        ::InternetCloseHandle(m_hSession) ;
        m_hSession == NULL ;
    }
}

CHTTPThread::~CHTTPThread()
{
    ResetSession() ;
}

//
//	Init(CHTTPThreadMgr *pINetThreadMgr, 
//       int iThreadNum, 
//       char *szTitle)
//

BOOL CHTTPThread::Init(CHTTPThreadMgr *pINetThreadMgr, 
                       int iThreadNum, 
                       char *szTitle)
{
    // Initialize the CHTTPThread, and let this instance
    // know about the CHTTPThreadMgr.

    ASSERT(pINetThreadMgr) ;

    m_pThreadMgr = pINetThreadMgr ;

    SetThreadNum(iThreadNum) ;

    strcpy(m_szTitle, szTitle) ;

    return (TRUE) ;
}

void CHTTPThread::SetThreadNum(int iThreadNum)
{
    m_iThreadNum = iThreadNum ;
}

void CHTTPThread::GetRemoteFile(CString& rURL, char *szLocalFile)
{
    ASSERT(!rURL.IsEmpty()) ;
    ASSERT(strlen(szLocalFile)) ;

    DebugMsg("CHTTPThread::GetRemoteFile()") ;

    strcpy(m_szRemoteFile, LPCTSTR(rURL)) ;

    //
    // Parse the http address for the server and file names
    //

    CString strTemp ;
	
    strTemp = rURL ;
    if (strTemp.Left(7) == "http://")
        strTemp = rURL.Mid(7) ; // Remove http://

    int i = strTemp.Find("/"); // Filename on address?
    if (i != -1)
    {
        m_strServer = strTemp.Left(i) ;
        m_strPath = strTemp.Mid(i) ;
    }
    else
        m_strServer = strTemp ;

    strcpy(m_szLocalFile, szLocalFile) ;

    //
    // Start a thread using AfxBeingThread() to get the file.
    //

    m_pThread = AfxBeginThread(GetRemoteFileThread, this) ;
}

//
// This is the thread function. 
//

UINT CHTTPThread::GetRemoteFileThread(LPVOID pvThread)
{
    // We need to get the current CHTTPThread instance
    // from the passed pointer to AfxBeginThread()

    CHTTPThread* pHTTPThread = (CHTTPThread*) pvThread ;

    if (pHTTPThread == NULL || 
        pHTTPThread->m_strServer.IsEmpty() || 
        pHTTPThread->m_pThreadMgr == NULL)
    {
        CHTTPThread::DebugMsg("CHTTPThread isn't a valid\r\n") ;
        return (HTTPTHREAD_FAILURE) ;
    }

    UINT uiRet = pHTTPThread->DownloadRemoteFile() ;

    // Reset the thread for next time...
    pHTTPThread->m_pThread = NULL ;

    // Thread terminates when this function returns...
    return (uiRet) ;
}

//
// This is where all of the actually Internet work is done.
//

UINT CHTTPThread::DownloadRemoteFile()
{
    UINT        uiResult = HTTPTHREAD_FAILURE ;
    char        *pBuffer ;
    HINTERNET   hConnection ;
    HINTERNET   hHttpFile ;
    BOOL        bSendRequest ;

    if (m_hSession == NULL)
    {
        //
        // Call InternetOpen() to get a handle
        // to an internet session.
        //

        m_hSession = ::InternetOpen(m_szTitle, 
                                    PRE_CONFIG_INTERNET_ACCESS, 
                                    NULL, 
                                    NULL, 
                                    0) ;

        if (!m_hSession)
        {
            CHTTPThread::DebugMsg("InternetOpen() failed") ;

            // Notify the thread manager that this thread failed...
            if (m_pThreadMgr)
                m_pThreadMgr->OnThreadComplete(m_iThreadNum, 
                    m_szLocalFile, HTTPTHREAD_FAILURE) ;

            return (HTTPTHREAD_FAILURE) ;
        }
    }

    hConnection = ::InternetConnect(m_hSession, 
                                    m_strServer, 
                                    INTERNET_DEFAULT_HTTP_PORT, 
                                    "", 
                                    "", 
                                    INTERNET_SERVICE_HTTP, 
                                    0, 
                                    0) ;

    if (hConnection == NULL)
    {
        CHTTPThread::DebugMsg("hConnection is NULL") ;
        uiResult = HTTPTHREAD_COULDNOTCONNECT ;
    }

    if (hConnection != NULL)
    {
        hHttpFile = ::HttpOpenRequest(hConnection, 
                                      "GET", 
                                      m_strPath, 
                                      HTTP_VERSION, 
                                      "", 
                                      0, 
                                      INTERNET_FLAG_RELOAD, 
                                      0) ;

        if (hHttpFile != NULL)
        {
            bSendRequest = ::HttpSendRequest(hHttpFile, 
                                             "", 
                                             0, 
                                             0, 
                                             0) ;
         
            if (!bSendRequest)
            {
                CHTTPThread::DebugMsg("HttpSendRequest failed") ;
                uiResult = HTTPTHREAD_FILENOTFOUND ;
            }
            else
            {
                // Get size of file.
                char	buffer[32] ;
                DWORD	dwFileSize ;
                DWORD	dwBufSize = sizeof(buffer) ;
                BOOL	bQuery ;

                bQuery = ::HttpQueryInfo(hHttpFile, 
                                         HTTP_QUERY_CONTENT_LENGTH, 
                                         buffer, 
                                         &dwBufSize, 
                                         0) ;

                if (bQuery)
                {
                    // HttpQueryInfo was successful, 
                    // so allocate the memory.
                    dwFileSize = (DWORD) atol(buffer) ;
                }
                else
                {
                    // HttpQueryInfo failed, but allocate
                    // an 8K block anyway...
                    dwFileSize = 8192 ;
                }

                pBuffer = new char[dwFileSize + 1] ;
				
                DWORD	dwTotalBytes = dwFileSize + 1 ;
                DWORD	dwBytesRead ;
                DWORD	dwTotalBytesRead = 0 ;

                dwTotalBytes = dwFileSize ;

                BOOL	bRead ;

                bRead = ::InternetReadFile(hHttpFile, 
                                           pBuffer, 
                                           dwTotalBytes, 
                                           &dwBytesRead) ;

                if (!bRead)
                {
                    // InternetReadFile failed...
                }
                else
                {
                    // Null-terminate the buffer
                    pBuffer[dwBytesRead] = 0 ;
                    uiResult = HTTPTHREAD_SUCCESS ;

                    // Save the data to the local file
                    if (strlen(m_szLocalFile))
                    {
                        CFile	theFile ;

                        if (theFile.Open(m_szLocalFile, 
                                         CFile::modeCreate | 
                                         CFile::modeWrite))
                        {
                            theFile.Write(pBuffer, dwBytesRead) ;
                            theFile.Close() ;
                        }
                        else
                        {
                            uiResult = HTTPTHREAD_COULDNOTSAVEFILE ;
                        }
                    }
                }

                // free the buffer
                delete [] pBuffer ;
            }
			::InternetCloseHandle(hHttpFile) ; 
        }

        CHTTPThread::DebugMsg("closing connection...") ;

        if (!::InternetCloseHandle(hConnection))
        {
            CHTTPThread::DebugMsg("InternetCloseHandle() failed") ;
        }
    }

    // Notify the thread manager that this thread is complete...
    if (m_pThreadMgr)
        m_pThreadMgr->OnThreadComplete(m_iThreadNum, 
            m_szLocalFile, uiResult) ;

    return (uiResult) ;
}

void CHTTPThread::DebugMsg(char *szMsg)
{
    // Simple debugging if we have compiled the DEBUG project

    #ifdef _DEBUG
    AfxMessageBox(szMsg) ;
    #endif
}

