//------------------------------------------------------------------
//  Kernprof.c - Analysis of kernprof.exe        Ton Plooy  aug 1997
//------------------------------------------------------------------
#define  _CRTIMP __declspec(dllimport)
#include <stdio.h>
#include <stdlib.h>
#define  WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <winioctl.h>
#include <imagehlp.h>
#include "winntddk.h"
#include "ntdll.h"


#define MAX_PROCESSORS                  32

#pragma pack(1)
typedef struct _QUERYSTAT {
    DWORD  dwXX0;                       //
    DWORD  dwXX1;                       //
    DWORD  dwBase;                      //
    DWORD  dwSize;                      //
    DWORD  dwXX2;                       //
    DWORD  dwXX3;                       //
    WORD   wXX4;                        //
    WORD   wModNameOffset;              // Base name offset 
    char   szName[256];                 // Module name and path
} QUERYSTAT;

typedef struct _MODTAB {                // Module Table
    HANDLE hProfile[MAX_PROCESSORS];    // Profile Handle
    DWORD  dwBase;                      // Image Base
    DWORD  dwBase2;                     // Image Base
    DWORD  dwSize;                      // Image size
    LPBYTE lpAlloc[MAX_PROCESSORS];     // Profile counters array
    DWORD  dwProfSize;                  // Profile size
    DWORD  dwInit;                      // 
    DWORD  nBucketSize;                 // Address granularity
    LPSTR  pModName;                    // Pointer to module name
    BOOL   bSymbols;                    // Do we have symbols
} MODTAB;

#pragma pack()


#define MAX_MODULES                 50  // Maximum number of modules

HANDLE _hBreakEvent;                    // Ctrl-C was hit
char   _szSrchPath[4096];               // Symbol search path
BOOL   _bPerProcessor   = FALSE;        // Per processor profiling
DWORD  _dwNumProcessors = 1;            // Number of processors
DWORD  _dwSource        = 0;            // Clock source
DWORD  _dwModCount      = 0;            // Global module count
MODTAB  ModTab[MAX_MODULES];            // Module table

//------------------------------------------------------------------
BYTE SymbolStructData[sizeof(IMAGEHLP_SYMBOL) + 1024];
BYTE SymbolStructCopy[sizeof(IMAGEHLP_SYMBOL) + 1024];
IMAGEHLP_SYMBOL *pSymbol = (IMAGEHLP_SYMBOL *)SymbolStructData;
IMAGEHLP_SYMBOL *pSymcpy = (IMAGEHLP_SYMBOL *)SymbolStructCopy;

//---- local prototypes --------------------------------------------
NTSTATUS _stdcall Initialize(void);
NTSTATUS _stdcall InitializeImage(HANDLE, QUERYSTAT *);
NTSTATUS _stdcall StartProfile(void);
NTSTATUS _stdcall StopProfile(void);
NTSTATUS _stdcall AnalyzeProfile(int);
void     _stdcall PrintEntries(MODTAB *, LPBYTE, int);
void     _stdcall Output(DWORD, MODTAB *, IMAGEHLP_SYMBOL *,
                         int, LPBYTE, LPBYTE);

//------------------------------------------------------------------

BOOL _stdcall ConsoleHandler(DWORD dwCtrlType)
{
    if (dwCtrlType != CTRL_C_EVENT)
        return FALSE;
    SetEvent(_hBreakEvent);
    return TRUE;
}

//------------------------------------------------------------------

void _stdcall PrintUsage()
{
    printf("Kernel Profiler Usage:\n\n");
    printf("Kernprof [-p] [<sample time> [<low threshold>]]\n");
    printf("      -p           - Per-processor profile objects\n");
    printf("   <sample time>   - Specify, in seconds, how long \n");
    printf("                     to collect profile information\n");
    printf("                     Default is wait until Ctrl-C\n");
    printf("   <low threshold> - Minimum number of counts to\n");
    printf("                     report. Defaults is 100\n");
    return ;
}

//------------------------------------------------------------------
//
//------------------------------------------------------------------

int main(int argc, char *argv[])
{
    long        lSampleTime      = -1;
    int         i, nLowThreshold = 100;
    NTSTATUS    ntStat;
    SYSTEM_INFO SysInfo;

    pSymbol->SizeOfStruct  = sizeof(IMAGEHLP_SYMBOL);
    pSymbol->MaxNameLength = 0x400;

    SymSetOptions(SYMOPT_UNDNAME | SYMOPT_CASE_INSENSITIVE);
    SymInitialize((HANDLE)-1, NULL, FALSE);
    SymGetSearchPath((HANDLE)-1, _szSrchPath, sizeof(_szSrchPath));

    _hBreakEvent = CreateEvent(NULL, FALSE, FALSE, NULL);

    if (argc > 1) {
        // Handle program arguments
        if (argv[1][1] == '?') {
            PrintUsage();
            return 0;
        }
        for (i = 1; i < argc; i++) {

            if (argv[i][0] == '-') {
                // Handle options. Only -p is handled here.
                switch (toupper(argv[i][1])) {

                case 'P':
                    GetSystemInfo(&SysInfo);
                    _bPerProcessor   = TRUE;
                    _dwNumProcessors = SysInfo.dwNumberOfProcessors;
                    break;

                default:
                    break;
                }
            }
            else {
                // Get arguments <sample time> <low threshold>
                if (lSampleTime != -1) {
                    nLowThreshold = atoi(argv[i]);
                }
                else {
                    lSampleTime = atoi(argv[i]);
                }
            }
        }
    }

    if ((ntStat = Initialize()) < 0) {
        printf("initialize failed status - %lx\n", ntStat);
        return ntStat;
    }

    SetConsoleCtrlHandler(ConsoleHandler, TRUE);

    if ((ntStat = StartProfile()) < 0) {
        printf("start profile failed status - %lx\n", ntStat);
        return ntStat;
    }

    if (lSampleTime == -1)
        printf("delaying until ^C\n");
    else
        printf("delaying for %ld seconds\n", lSampleTime);

    if (lSampleTime == 0L) {
        getchar();
    }
    else {

        if (lSampleTime != -1L) {
            lSampleTime *= 1000;
        }
        if (_hBreakEvent == NULL)
            Sleep(lSampleTime);
        else
            WaitForSingleObject(_hBreakEvent, lSampleTime);
    }
    printf("end of delay\n");

    if ((ntStat = StopProfile()) < 0) {
        printf("stop profile failed status - %lx\n", ntStat);
        return ntStat;
    }

    SetConsoleCtrlHandler(ConsoleHandler, FALSE);

    AnalyzeProfile(nLowThreshold);

    return 0;
}

//------------------------------------------------------------------

NTSTATUS _stdcall Initialize()
{
    BYTE           Priv1, Priv2;        // Privilege tokens
    NTSTATUS       ntStat = 0;          // Status code
    UNICODE_STRING ucR, ucRSD, ucRS;    // Not used by program?
    DWORD          i, dwSize = 0, dwNum;// Generic counters
    char           Data[64000];         // NtQuery buffer
    QUERYSTAT     *pStat;               // Query buffer pointer
    HANDLE         hImage;              // Exe image handle
    char           szImage[256];        // Exe file name

    // Get information on loaded system modules
    if ((ntStat = NtQuerySystemInformation(0x0B, Data,
                                    sizeof(Data), &dwSize)) != 0) {
        printf("query system info failed - status %08lX\n", ntStat);
        return ntStat;
    }

    RtlInitUnicodeString(&ucR  , L"\\SystemRoot\\");
    RtlInitUnicodeString(&ucRSD,
                        L"\\SystemRoot\\System32\\Drivers\\");
    RtlInitUnicodeString(&ucRS , L"\\SystemRoot\\System32\\");

    ntStat = RtlAdjustPrivilege(SE_SYSTEM_PROFILE_PRIVILEGE,
                                                TRUE, 0, &Priv2);
    if ((ntStat != 0) && (ntStat != STATUS_NOT_ALL_ASSIGNED)) {
        printf("Enable system profile privilege "
               "failed - status 0x%lX\n", ntStat);
    }
    ntStat = RtlAdjustPrivilege(SE_INCREASE_QUOTA_PRIVILEGE,
                                                TRUE, 0, &Priv1);
    if ((ntStat != 0) && (ntStat != STATUS_NOT_ALL_ASSIGNED)) {
        printf("Unable to increase quota privilege "
               "(status=0x%lx)\n", ntStat);
    }

    dwNum = *(DWORD *)Data;
    pStat = (QUERYSTAT *)(Data + 4);
    for (i = 0; (i < dwNum) && (_dwModCount < MAX_MODULES); i++) {
        // Find image symbol information
        if (SymLoadModule((HANDLE)-1, NULL,
                      pStat->szName + pStat->wModNameOffset, NULL,
                      pStat->dwBase, pStat->dwSize) != 0L) {
            ModTab[i].bSymbols = TRUE;
        }
        else {
            ModTab[i].bSymbols = FALSE;
        }
        // Get image handle
        if ((hImage = FindExecutableImage(pStat->szName +
                                pStat->wModNameOffset,
                                _szSrchPath, szImage)) != NULL) {
            ntStat = InitializeImage(hImage, pStat);
            CloseHandle(hImage);
        }
        _dwModCount++;
        pStat++;
    }

    if (Priv2 == 0) {
        ntStat = RtlAdjustPrivilege(SE_SYSTEM_PROFILE_PRIVILEGE,
                                                FALSE, 0, &Priv2);
        if ((ntStat < 0) || (ntStat == STATUS_NOT_ALL_ASSIGNED))
            printf("Disable system profile privilege "
                   "failed - status 0x%lx\n", ntStat);
    }
    if (Priv1 == 0) {
        // NB: code has a bug here. Instead of quota, system is 
        // used again
        ntStat = RtlAdjustPrivilege(SE_INCREASE_QUOTA_PRIVILEGE,
                                                FALSE, 0, &Priv1);
        if ((ntStat < 0) || (ntStat == STATUS_NOT_ALL_ASSIGNED))
            printf("Disable increate quota privilege "
                   "failed - status 0x%lx\n", ntStat);
    }
    return ntStat;
}

//------------------------------------------------------------------

NTSTATUS _stdcall InitializeImage(HANDLE hImage, QUERYSTAT *pStat)
{
    int               i, nCPU;
    HANDLE            hMapping;
    LPBYTE            pFileMap;
    IMAGE_NT_HEADERS *pNtHdr;
    DWORD             dwSize, dwAllocSize, dwXX;
    DWORD             dwAddress;
    NTSTATUS          ntStat;
    QUOTA_LIMITS      QuotaLimits;

    if ((hMapping = CreateFileMapping(hImage, NULL, PAGE_READONLY,
                                           0L, 0L, NULL)) == NULL) {
        return 0;
    }
    if ((pFileMap = MapViewOfFile(hMapping, FILE_MAP_READ,
                                             0L, 0L, 0L)) == NULL) {
        CloseHandle(hMapping);
        return 0;
    }
    pNtHdr = ImageNtHeader(pFileMap);

    // The SizeOfImage is the total size of all sections, e.g. the
    // size off the image that the loader has to handle. It is very
    // much dependent on the SectionAlignment value. Each section 
    // size is rounded up to a page size and this is rounded up to 
    // the SectionAlignment size. The sum of these sizes for all 
    // sections is the SizeOfImage.
    dwSize = pNtHdr->OptionalHeader.SizeOfImage;
    if ((pNtHdr->FileHeader.Machine == IMAGE_FILE_MACHINE_R4000) || 
        (pNtHdr->FileHeader.Machine == IMAGE_FILE_MACHINE_R10000)) {
            // Apparently relocations don't exist on these machines 
            dwSize -= pNtHdr->OptionalHeader.
                DataDirectory[IMAGE_DIRECTORY_ENTRY_BASERELOC].Size;
    }
    UnmapViewOfFile(pFileMap);
    CloseHandle(hMapping);

    // We allocate a max of 512K or dwSize + one page
    dwAllocSize = (dwSize > 0x80000L) ? 0x80000L : dwSize + 0x1000;

    ModTab[_dwModCount].dwBase   = pStat->dwBase;
    ModTab[_dwModCount].dwBase2  = pStat->dwBase;
    ModTab[_dwModCount].dwSize   = dwSize;          // Image size
    ModTab[_dwModCount].dwInit   = TRUE;
    ModTab[_dwModCount].pModName =
                    strdup(pStat->szName + pStat->wModNameOffset);

    for (nCPU = 0; nCPU < (int)_dwNumProcessors; nCPU++) {
        dwAddress = 0;
        if ((ntStat = NtAllocateVirtualMemory((HANDLE)-1,
                            &dwAddress, 0L,
                            &dwAllocSize, MEM_COMMIT | MEM_RESERVE,
                            PAGE_READWRITE)) < 0) {
            printf("alloc VM failed %lx\n", ntStat);
            return ntStat;
        }
        // Calculate bucket size, start with a two-byte bucket. If 
        // the image is larger than the allocated size, we get a 
        // bigger bucket. The sample granularity is 2^n where n is
        // the bucket size. A default bucket size of 2 means that 
        // every 4-bytes are sampled.
        i = 2;
        dwXX = (dwSize / (dwAllocSize / 4) / 4); // Align
        while (dwXX != 0) {
            dwXX /= 2;
            i++;
        }

        ModTab[_dwModCount].lpAlloc[nCPU] = (LPBYTE)dwAddress;
        ModTab[_dwModCount].dwProfSize  = dwSize >> (i - 2);
        ModTab[_dwModCount].nBucketSize = i;

        if ((ntStat = NtQueryInformationProcess((HANDLE)-1,
                            ProcessQuotaLimits, &QuotaLimits,
                            sizeof(QuotaLimits), NULL)) < 0) {
            printf("query process info failed %lx\n", ntStat);
            return ntStat;
        }
        QuotaLimits.MaximumWorkingSetSize += dwAllocSize;
        QuotaLimits.MinimumWorkingSetSize += dwAllocSize;
        NtSetInformationProcess((HANDLE)-1, ProcessQuotaLimits,
                            &QuotaLimits, sizeof(QuotaLimits));

        if ((ntStat = NtCreateProfile(
                            &ModTab[_dwModCount].hProfile[nCPU],
                            0,
                            ModTab[_dwModCount].dwBase2,
                            ModTab[_dwModCount].dwSize,
                            ModTab[_dwModCount].nBucketSize,
                            ModTab[_dwModCount].lpAlloc[nCPU],
                            ModTab[_dwModCount].dwProfSize,
                            _dwSource,
                            (_bPerProcessor == FALSE) ?
                                    (DWORD)-1 : 1 << nCPU)) != 0) {
            printf("create kernel profile %s failed - status %lx\n",
                    ModTab[_dwModCount].pModName, ntStat);
        }
    }
    return ntStat;
}

//------------------------------------------------------------------

NTSTATUS _stdcall StartProfile()
{
    int          i, j, k = 0;
    NTSTATUS     ntStat;
    QUOTA_LIMITS QuotaLimits;

    NtSetIntervalProfile(10000, _dwSource);

    for (i = 0; i < (int)_dwModCount; i++) {

        for (j = 0; j < (int)_dwNumProcessors; j++) {

            ntStat = NtStartProfile(ModTab[i].hProfile[j]);
            if (ntStat == STATUS_WORKING_SET_QUOTA) {

                if ((ntStat = NtQueryInformationProcess((HANDLE)-1,
                                ProcessQuotaLimits, &QuotaLimits,
                                sizeof(QuotaLimits), NULL)) < 0) {
                    printf("query process info "
                           "failed %lx\n", ntStat);
                    return ntStat;
                }
                QuotaLimits.MinimumWorkingSetSize +=
                                ModTab[i].dwProfSize + 0x14000;
                QuotaLimits.MaximumWorkingSetSize +=
                                ModTab[i].dwProfSize + 0x14000;
                ntStat = NtSetInformationProcess((HANDLE)-1,
                                ProcessQuotaLimits,
                                &QuotaLimits, sizeof(QuotaLimits));

                ntStat = NtStartProfile(ModTab[i].hProfile[j]);
            }
            if (ntStat < 0) {
                printf("start profile %s failed - status %lx\n",
                        ModTab[i].pModName, ntStat);
            }
        }
    }
    return ntStat;
}

//------------------------------------------------------------------

NTSTATUS _stdcall StopProfile()
{
    int      i, j;
    NTSTATUS ntStat;

    for (i = 0; i < (int)_dwModCount; i++) {

        for (j = 0; j < (int)_dwNumProcessors; j++) {

            ntStat = NtStopProfile(ModTab[i].hProfile[j]);
            if (ntStat != 0) {
                printf("stop profile %s failed - status %lx\n",
                        ModTab[i].pModName, ntStat);
            }
        }
    }
    return 0;
}

//------------------------------------------------------------------

NTSTATUS _stdcall AnalyzeProfile(int nLow)
{
    int    i, j;
    DWORD  dwCount, dwTotal = 0;
    LPBYTE pData, pEnd;

    // NB: code repeats stop profile code here, we skip that

    for (i = 0; i < (int)_dwNumProcessors; i++) {

        if (_bPerProcessor == TRUE)
            printf("\nPROCESSOR %d\n", i);

        for (j = 0; j < (int)_dwModCount; j++) {
            
            pData   = ModTab[j].lpAlloc[i];
            pEnd    = pData + (ModTab[j].dwProfSize & 0xFFFFFFFC);
            dwCount = 0;
            while (pData < pEnd) {
                pEnd    -= 4;
                dwCount += *(DWORD *)pEnd;  // Totalize
            }
            dwTotal += dwCount;

            if (dwCount >= (DWORD)nLow) {

                printf("\n%9d %20s --Total Hits-- %s\n",
                        dwCount, ModTab[j].pModName,
                        (ModTab[j].bSymbols) ? "" : "(NO SYMBOLS)");

                if (ModTab[j].bSymbols == TRUE) {
                    PrintEntries(&ModTab[j], pData, nLow);
                }
            }
        }
    }
    return 0;
}

//------------------------------------------------------------------

void _stdcall PrintEntries(MODTAB *pModTab, LPBYTE pData, int nLow)
{
    DWORD   dwAddress, dwDisplacement, dwCount = 0;
    LPBYTE  pInfo, pEnd, pCount;
    BOOL    bCopy = FALSE;

    pInfo = pData;
    pEnd  = pData + (pModTab->dwProfSize & 0xFFFFFFFC);
    for (pCount = pData; pCount < pEnd; pCount += 4) {

        if (*(DWORD *)pCount != 0) {

            dwAddress = pModTab->dwBase2 +
                ((pCount - pData) << (pModTab->nBucketSize - 2));

            if (SymGetSymFromAddr((HANDLE)-1, dwAddress,
                            &dwDisplacement, pSymbol) == TRUE) {
                // We have a symbol name, start adding hit counts
                if ((bCopy == TRUE) && (pSymcpy->Address != 0) &&
                        (pSymbol->Address == pSymcpy->Address)) {
                    dwCount += *(DWORD *)pCount;
                }
                else {
                    Output(dwCount, pModTab, pSymcpy,
                                            nLow, pInfo, pCount);
                    pInfo   = pCount;
                    dwCount = *(DWORD *)pCount;
                    memcpy(pSymcpy, pSymbol,
                            sizeof(SymbolStructData));
                    bCopy = TRUE;
                }
            }
            else {
                Output(dwCount, pModTab, pSymcpy,
                                            nLow, pInfo, pCount);
            }
            
        }
    }
    // Print last symbol
    Output(dwCount, pModTab, pSymcpy, nLow, pInfo, pCount);

    return ;
}

//------------------------------------------------------------------

void _stdcall Output(DWORD dwCount, MODTAB *pModTab,
                     IMAGEHLP_SYMBOL *pSymbol,
                     int nLowThreshold, LPBYTE pData, LPBYTE pCount)
{
    if (dwCount < (DWORD)nLowThreshold)
        return ;

    printf("%9ld ", dwCount);
    printf("%20s %s ", pModTab->pModName, pSymbol->Name);
    printf("\n");
    return ;
}

